from __future__ import annotations

from abc import abstractmethod
from math import floor
from typing import TYPE_CHECKING
from typing import Generic
from typing import TypeVar
from typing import cast

from textual.geometry import NULL_SPACING
from textual.geometry import Offset
from textual.geometry import Region
from textual.geometry import Size
from textual.layout import ArrangeResult
from textual.layout import Layout
from textual.layout import WidgetPlacement
from textual.layouts.horizontal import HorizontalLayout
from textual.layouts.vertical import VerticalLayout

from ._timeline_entry import AbstractEntry
from ._timeline_entry import HorizontalEntry
from ._timeline_entry import VerticalEntry

if TYPE_CHECKING:
    from collections.abc import Callable

    from textual.css.scalar import Scalar

    from ._base_timeline import AbstractTimeline


T = TypeVar("T", bound=AbstractEntry)


class AbstractTimelineLayout(Layout, Generic[T]):
    SubLayout: type[Layout]
    name = "timeline"

    def __init__(self, *, tile: bool = True) -> None:
        self.tile = tile
        self._layout = self.SubLayout()

    def arrange(  # type: ignore[override] # NOTE: Always will be attached to a timeline widget.
        self,
        parent: AbstractTimeline[T],
        entries: list[T],
        size: Size,
        *,
        greedy: bool = True,
    ) -> ArrangeResult:
        if not self.tile:
            return self._layout.arrange(parent, entries, size, greedy=greedy)  # type: ignore[arg-type] # NOTE: Asking for a parent class anyway.

        parent.pre_layout(self)
        # NOTE: Sorts the widgets in pre-layout by offset and size.

        return [
            WidgetPlacement(
                region=region,
                offset=offset,
                margin=NULL_SPACING,
                widget=entry,
                absolute=True,
            )
            for entry, (region, offset) in zip(
                entries, self._tile_entries(size, entries), strict=False
            )
        ]

    def _process_group(
        self,
        add_tile: Callable[[tuple[int, int]], None],
        slots: list[T],
        size: int,
    ) -> None:
        n = len(slots)
        tile_size = round(size / n)

        for i in range(n):
            add_tile((i * tile_size, tile_size))

    @abstractmethod
    def _tile_entries(
        self,
        size: Size,
        entries: list[T],
    ) -> list[tuple[Region, Offset]]: ...


class VerticalTimelineLayout(AbstractTimelineLayout[VerticalEntry]):
    SubLayout = VerticalLayout
    name = "vertical_timeline"

    def _tile_entries(
        self,
        size: Size,
        entries: list[VerticalEntry],
    ) -> list[tuple[Region, Offset]]:
        # NOTE: Uses this as reference: https://github.com/rajeshdavidbabu/Javascript-Event-Calender
        previous_end = None
        columns = list[VerticalEntry]()
        add_column = columns.append
        tiling = list[tuple[int, int]]()
        add_tile = tiling.append
        for entry in entries:
            if previous_end is not None and entry.offset.y >= previous_end:
                self._process_group(add_tile, columns, size.width)
                previous_end = None
                columns.clear()

            add_column(entry)

            end = entry.offset.y + cast("Scalar", entry.styles.height).value
            if previous_end is None or end > previous_end:
                previous_end = end

        if columns:
            self._process_group(add_tile, columns, size.width)

        return [
            (
                Region(
                    0,
                    0,
                    tile_width,
                    round(cast("Scalar", entry.styles.height).value),
                ),
                Offset(col, floor(entry.offset.y)),
            )
            for entry, (col, tile_width) in zip(entries, tiling, strict=False)
        ]


class HorizontalTimelineLayout(AbstractTimelineLayout[HorizontalEntry]):
    SubLayout = HorizontalLayout
    name = "horizontal_timeline"

    def _tile_entries(
        self,
        size: Size,
        entries: list[HorizontalEntry],
    ) -> list[tuple[Region, Offset]]:
        # NOTE: Uses this as reference: https://github.com/rajeshdavidbabu/Javascript-Event-Calender
        previous_end = None
        rows = list[HorizontalEntry]()
        add_row = rows.append
        tiling = list[tuple[int, int]]()
        add_tile = tiling.append
        for entry in entries:
            if previous_end is not None and entry.offset.x >= previous_end:
                self._process_group(add_tile, rows, size.height)
                previous_end = None
                rows.clear()

            add_row(entry)

            end = entry.offset.x + cast("Scalar", entry.styles.width).value
            if previous_end is None or end > previous_end:
                previous_end = end

        if rows:
            self._process_group(add_tile, rows, size.height)

        return [
            (
                Region(
                    0,
                    0,
                    round(cast("Scalar", entry.styles.width).value),
                    tile_height,
                ),
                Offset(floor(entry.offset.x), row),
            )
            for entry, (row, tile_height) in zip(entries, tiling, strict=False)
        ]



"""Holds mostly composite widgets that build upon the timeline."""

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import ClassVar
from typing import Generic
from typing import TypeVar
from typing import cast

from textual.containers import ScrollableContainer
from textual.css.query import NoMatches
from textual.reactive import reactive
from textual.reactive import var

from textual_timepiece._utility import format_seconds
from textual_timepiece.timeline._base_timeline import HorizontalRuler
from textual_timepiece.timeline._base_timeline import VerticalRuler

from ._base_timeline import AbstractRuler
from ._base_timeline import HorizontalEntryType
from ._base_timeline import HorizontalTimelineNavigation
from ._base_timeline import TimelineNavigation
from ._base_timeline import VerticalEntryType
from ._base_timeline import VerticalTimelineNavigation

if TYPE_CHECKING:
    from textual.app import ComposeResult
    from textual.widget import Widget


def _default_marker_factory(index: int) -> str:
    return format_seconds(index, include_seconds=False)


Navigation = TypeVar("Navigation", bound=TimelineNavigation[Any])
Ruler = TypeVar("Ruler", bound=AbstractRuler)


class AbstractRuledTimeline(ScrollableContainer, Generic[Navigation, Ruler]):
    """Ruled timelines with markers.

    Params:
        total: Total amount of timelines to draw.
        marker_factory: Factory function for creating markers on the ruler.
            Defaults to generating markers in the *HH:mm* format.
        header_factory: Factory function for creating headers.
        name: The name of the widget.
        id: The ID of the widget in the DOM.
        classes: The CSS classes for the widget.
        disabled: Whether the widget is disabled or not.
        can_maximize: Whether the widget can be maximized or not.
    """

    Timeline: type[Navigation]

    total = var[int](1, init=False)
    """Total amount of timelines to compose. Minimum 1."""

    duration = var[int](86400, init=False)
    """Total time included in widget in seconds."""

    length = reactive[int](96, init=False, layout=True)
    """Actual length of the timelines."""

    def __init__(
        self,
        total: int | None = None,
        marker_factory: AbstractRuler.MarkerFactory | None = None,
        header_factory: Callable[[int], Widget] | None = None,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        *,
        disabled: bool = False,
        can_maximize: bool | None = None,
    ) -> None:
        super().__init__(
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
            can_maximize=can_maximize,
        )
        if total:
            self.set_reactive(
                AbstractRuledTimeline.total,
                self._validate_total(total),
            )

        self._marker_factory = marker_factory or _default_marker_factory
        self._header_factory = header_factory

    def compose(self) -> ComposeResult:
        for i in range(1, self.total + 1):
            header = self._header_factory(i) if self._header_factory else None
            yield self.Timeline(header, id=f"timeline-{i}").data_bind(
                AbstractRuledTimeline.length
            )

    def _validate_total(self, value: int) -> int:
        return max(value, 1)

    async def _watch_total(self, old: int, new: int) -> None:
        if old < new:
            await self.mount(
                *(
                    self.Timeline(
                        self._header_factory(i)
                        if self._header_factory
                        else None,
                        id=f"timeline-{i}",
                    ).data_bind(AbstractRuledTimeline.length)
                    for i in range(old + 1, new + 1)
                )
            )
        else:
            timelines: list[Navigation] = []
            for i in range(old, new, -1):
                try:
                    tl = self.query_one(f"#timeline-{i}", self.Timeline)
                except NoMatches as err:
                    self.log.debug(err)
                else:
                    timelines.append(tl)
            await self.remove_children(timelines)

    @property
    def ruler(self) -> Ruler:
        return cast("Ruler", self.query_exactly_one(AbstractRuler))

    @property
    def subdivisions(self) -> int:
        return self.ruler.subdivisions

    @subdivisions.setter
    def subdivisions(self, value: int) -> None:
        self.ruler.subdivisions = value


class RuledVerticalTimeline(
    AbstractRuledTimeline[VerticalTimelineNavigation, VerticalRuler]
):
    """Ruled vertical timeline with markers.

    !!! note
        If providing headers with the `header_factory` parameter make
        sure to compensate with top padding for the ruler to keep alignment
        in place.

    Params:
        total: Total amount of timelines to draw.
        marker_factory: Factory function for creating markers on the ruler.
            Defaults to generating markers in the *HH:mm* format.
        header_factory: Factory function for creating headers.
        name: The name of the widget.
        id: The ID of the widget in the DOM.
        classes: The CSS classes for the widget.
        disabled: Whether the widget is disabled or not.
        can_maximize: Whether the widget can be maximized or not.
    """

    Timeline = VerticalTimelineNavigation[VerticalEntryType]
    DEFAULT_CSS: ClassVar[str] = """\
    RuledVerticalTimeline {
        layout: horizontal;
        width: 1fr;
        & > VerticalTimelineNavigation {
            width: 1fr;
            & > VerticalTimeline {
                width: 1fr;
            }
        }
    }
    """
    """Default CSS for `RuledVerticalTimeline` widget."""

    def compose(self) -> ComposeResult:
        yield VerticalRuler(marker_factory=self._marker_factory).data_bind(
            RuledVerticalTimeline.length
        )
        yield from super().compose()


class RuledHorizontalTimeline(
    AbstractRuledTimeline[HorizontalTimelineNavigation, HorizontalRuler]
):
    """Ruled horizontal timeline with markers.

    !!! note
        If providing headers with the `header_factory` parameter make
        sure to compensate with left padding for the ruler to keep alignment
        in place.

    Params:
        total: Total amount of timelines to draw.
        marker_factory: Factory function for creating markers on the ruler.
            Defaults to generating markers in the *HH:mm* format.
        header_factory: Factory function for creating headers.

        name: The name of the widget.
        id: The ID of the widget in the DOM.
        classes: The CSS classes for the widget.
        disabled: Whether the widget is disabled or not.
        can_maximize: Whether the widget can be maximized or not.
    """

    Timeline = HorizontalTimelineNavigation[HorizontalEntryType]
    DEFAULT_CSS: ClassVar[str] = """\
    RuledHorizontalTimeline {
        height: 1fr;
        & > HorizontalTimelineNavigation {
            height: 1fr;
            & > HorizontalTimeline {
                height: 1fr;
            }
        }
    }
    """
    """Default CSS for `RuledHorizontalTimeline` widget."""

    def __init__(
        self,
        total: int | None = None,
        marker_factory: AbstractRuler.MarkerFactory | None = None,
        header_factory: Callable[[int], Widget] | None = None,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        *,
        disabled: bool = False,
        can_maximize: bool | None = None,
    ) -> None:
        super().__init__(
            total,
            marker_factory,
            header_factory,
            name,
            id,
            classes,
            disabled=disabled,
            can_maximize=can_maximize,
        )
        self.length = 192

    def compose(self) -> ComposeResult:
        yield HorizontalRuler(marker_factory=self._marker_factory).data_bind(
            RuledHorizontalTimeline.length
        )
        yield from super().compose()



from __future__ import annotations

import sys
from abc import abstractmethod
from typing import TYPE_CHECKING
from typing import ClassVar
from typing import TypeVar
from typing import cast
from uuid import uuid4

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from textual.geometry import Offset
from textual.geometry import Size
from textual.reactive import var
from textual.widgets import Static

from textual_timepiece._extra import BaseMessage

if TYPE_CHECKING:
    from rich.console import RenderableType
    from textual.await_remove import AwaitRemove
    from textual.css.scalar import Scalar
    from textual.dom import DOMNode
    from textual.events import Leave
    from textual.events import MouseDown
    from textual.events import MouseEvent
    from textual.events import MouseMove
    from textual.events import MouseUp
    from textual.visual import SupportsVisual

    from textual_timepiece.timeline._base_timeline import VerticalTimeline


T = TypeVar("T", bound="DOMNode")


class AbstractEntry(Static, can_focus=True):
    """Abstract entry widget for visual timeline widgets.

    Params:
        id: ID of the Widget.
        content: A Rich renderable, or string containing console markup.
        offset: Initial offset to use for the entry.
        size: Initial size to use for the entry.
        markup: True if markup should be parsed and rendered.
        name: Name of widget.
        classes: Space separated list of class names.
        disabled: Whether the static is disabled or not.
    """

    class Updated(BaseMessage[T]):
        """Base message for all entries."""

        @property
        def entry(self) -> T:
            return self.widget

    class Resized(Updated[T]):
        """Entry was resized by the user."""

        def __init__(self, widget: T, size: Size, delta: int) -> None:
            super().__init__(widget)
            self.size = size
            self.delta = delta

    class Moved(Updated[T]):
        """Entry was moved by the user."""

        def __init__(self, widget: T, delta: int) -> None:
            super().__init__(widget)
            self.delta = delta

    DIMENSION: ClassVar[str]
    DEFAULT_CSS: ClassVar[str] = """\
    AbstractEntry {
        box-sizing: content-box;
        position: absolute !important;
        background: $secondary-background-darken-2;
        content-align: center middle;
        text-align: center;
        color: auto;
        &:hover {
            border: thick $secondary;
        }
        &.selected {
            border: double $primary;
            color: $primary;
        }
        &:focus {
            color: $primary;
            text-style: bold;
        }
        &.moving {
            border: thick $primary;
        }
        &.-mime {
            opacity: 75%;
            color: auto;
        }
    }
    """

    EDGE_MARGIN: ClassVar[int]
    """The amount of space the margin takes up for the resizing grip."""

    ALLOW_MAXIMIZE = False

    clicked = var[Offset | None](None, init=False)
    """Initial point where the entry was clicked when moving or resizing."""

    _is_moving: bool | None = None
    _start_is_hovered: bool | None = None

    def __init__(
        self,
        id: str,
        content: RenderableType | SupportsVisual = "",
        offset: int | None = None,
        size: int | None = None,
        *,
        markup: bool = True,
        name: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ) -> None:
        super().__init__(
            content,
            markup=markup,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
            expand=False,
            shrink=False,
        )
        self.set_dims(offset, size)

    @classmethod
    def mime(
        cls,
        offset: int | None = None,
        size: int | None = None,
        *,
        name: str | None = None,
        markup: bool = True,
        disabled: bool = False,
    ) -> Self:
        """Instatiate a mime entry. For previewing resizing visually.

        Args:
            offset: Initial offset to use for the entry.
            size: Initial size to use for the entry.
            markup: True if markup should be parsed and rendered.
            name: Name of widget.
            disabled: Whether the static is disabled or not.

        Returns:
            Constructed timeline entry.
        """
        return cls(
            id=f"id-{uuid4()}",
            content="",
            offset=offset,
            size=size,
            markup=markup,
            name=name,
            classes="-mime",
            disabled=disabled,
        )

    async def _on_mouse_down(self, event: MouseDown) -> None:
        if self.app.mouse_captured is None:
            self.capture_mouse()
            self._is_focused(event.offset)

    def _on_mouse_move(self, event: MouseMove) -> None:
        if self.clicked:
            self._adjust(event)
        elif self.mouse_hover:
            self._add_edge(event.offset)
            self.add_class("hovered")

    async def _on_mouse_up(self, event: MouseUp) -> None:
        if self.app.mouse_captured == self:
            self.capture_mouse(False)
            await self._is_unfocused(event)

    def _adjust(self, event: MouseMove) -> None:
        if (
            self.clicked is not None
            and event.button != 0
            and self._is_moving is not None
        ):
            if self._is_moving:
                self.add_class("moving")
                self.move(self.sel_delta(event))
            elif self._start_is_hovered is not None:
                self.resize(
                    self.sel_delta(event),
                    tail=self._start_is_hovered,
                )

    def _on_leave(self, event: Leave) -> None:
        self.remove_class("size_end", "size_start")

    def _is_focused(self, offset: Offset) -> None:
        self._is_moving = self.is_moving(offset)
        if not self._is_moving:
            self._add_edge(offset)
            self._start_is_hovered = self.is_start(offset)

        self.clicked = offset

    async def _is_unfocused(self, event: MouseUp) -> None:
        self.clicked = None
        self._is_moving = None
        self._start_is_hovered = None
        self.remove_class("moving", "size_start", "size_end")

    def resize(self, delta: int, *, tail: bool) -> None:
        """Public method for resizing the widget.

        Args:
            delta: total amount to be moved.
            tail: Whether to adjust the end or the start of the widget.
        """
        delta = self._resize_helper(delta, tail=tail)
        self.post_message(self.Resized(self, self.size, delta))

    @abstractmethod
    def move(self, delta: int) -> None:
        """Public method for moving the entry.

        Args:
            delta: total amount to move the entry by.
        """

    @abstractmethod
    def _resize_helper(self, delta: int, *, tail: bool) -> int:
        """Resizing logic for the entry.

        Args:
            delta: Total amount to be moved.
            tail: Whether to adjust the end of the widget or the start.

        Returns:
            The total amount adjusted.
        """

    @abstractmethod
    def is_tail(self, offset: Offset) -> bool:
        """Is the offset hovering the tail of the widget?

        Args:
            offset: Mouse offset to the check against.

        Returns:
            True if its the tail else False.
        """

    @abstractmethod
    def is_head(self, offset: Offset) -> bool:
        """Is the offset hovering the head of the widget?

        Args:
            offset: Mouse offset to the check against.

        Returns:
            True if its the tail else False.
        """

    @abstractmethod
    def is_start(self, offset: Offset) -> bool:
        """Is the offset hovering the first half of the widget?

        Args:
            offset: Mouse offset to check against.

        Returns:
            True if the offset hovering the first half of the widget.
        """

    def is_moving(self, offset: Offset) -> bool:
        """Is the widget moving or resizing?

        Args:
            offset: Mouse offset to check against.

        Returns:
            True if the widget is moving else False.
        """
        return not (self.is_tail(offset) or self.is_head(offset))

    @abstractmethod
    def sel_delta(self, event: MouseEvent) -> int:
        """Select the correct delta for move/resize operations.

        Args:
            event: Mouse event to extract the delta from.

        Returns:
            The main delta of one mouse direction.
        """

    def _add_edge(self, offset: Offset) -> None:
        if self.clicked:
            # NOTE: Making sure the edge doesn't change after locking in.
            return

        self.set_class(self.is_tail(offset), "size_start")
        self.set_class(self.is_head(offset), "size_end")

    @abstractmethod
    def set_dims(
        self,
        offset: int | None = None,
        size: int | None = None,
    ) -> None:
        """Set entry dimensions directly.

        Args:
            offset: The offset of the entry in the core location.
            size: The size of the entry.
        """

    @abstractmethod
    def merge(self, other: Self) -> AwaitRemove:
        """Merge two entries into one taking the furthest extents.

        Args:
            other: The other entry to merge into.

        Returns:
            An `AwaitRemove` object for removing the supplied entry.
        """

    @property
    @abstractmethod
    def start(self) -> int:
        """The start of the entry."""

    @property
    def dimension(self) -> int:
        """Alias for the size of the entry depending on orientation."""
        return cast("int", getattr(self.styles, self.DIMENSION).value)

    @dimension.setter
    def dimension(self, value: int) -> None:
        """Set the size of the entry depening on orientation."""
        setattr(self.styles, self.DIMENSION, value)

    @property
    def end(self) -> int:
        return self.start + self.dimension


class VerticalEntry(AbstractEntry):
    """Vertical entry for a vertical timeline layout.

    Params:
        id: ID of the Widget.
        content: A Rich renderable, or string containing console markup.
        offset: Initial offset to use for the entry.
        size: Initial size to use for the entry.
        markup: True if markup should be parsed and rendered.
        name: Name of widget.
        classes: Space separated list of class names.
        disabled: Whether the static is disabled or not.
    """

    class Resized(AbstractEntry.Resized["VerticalEntry"]):
        """Entry was resized by the user."""

    class Moved(AbstractEntry.Moved["VerticalEntry"]):
        """Entry was moved by the user."""

    parent: VerticalTimeline
    DIMENSION = "height"
    EDGE_MARGIN: ClassVar[int] = 1
    DEFAULT_CSS: ClassVar[str] = """\
    VerticalEntry {
        width: 100%;
        min-height: 2;
        height: 2;
        border-top: double $panel-lighten-2;
        border-bottom: double $panel-lighten-2;
        &:hover.size_start {
            border: none;
            border-top: double $secondary;
            border-bottom: double $panel-lighten-2;
        }
        &:hover.size_end {
            border: none;
            border-bottom: double $secondary;
            border-top: double $panel-lighten-2;
        }
        &:focus.size_start {
            border: none;
            border-top: thick $primary;
            border-bottom: double $primary;
        }
        &:focus.size_end {
            border: none;
            border-bottom: thick $primary;
            border-top: double $primary;
        }
        &:focus {
            border: double $primary;
        }
        &.-mime {
            hatch: horizontal white 5%;
        }
    }
    """
    """Default CSS for `VerticalEntry` widget."""

    def sel_delta(self, event: MouseEvent) -> int:
        return event.delta_y

    def is_tail(self, offset: Offset) -> bool:
        return 0 <= offset.y <= self.EDGE_MARGIN

    def is_head(self, offset: Offset) -> bool:
        height = cast("Scalar", self.styles.height).value
        return height - self.EDGE_MARGIN <= offset.y <= height

    def is_start(self, offset: Offset) -> bool:
        return offset.y < cast("Scalar", self.styles.height).value / 2

    def move(self, delta: int) -> None:
        offset = Offset(0, delta)
        self.offset = self.offset + offset
        self.post_message(self.Moved(self, delta))

    def merge(self, other: VerticalEntry) -> AwaitRemove:
        y1, y2 = self.offset.y, other.offset.y
        start = y1 if y1 <= y2 else y2
        self.offset = Offset(0, start)
        e1, e2 = self.end, other.end
        self.styles.height = abs(start - (e1 if e1 >= e2 else e2))
        return other.remove()

    def _resize_helper(self, delta: int, *, tail: bool) -> int:
        if tail:
            delta *= -1

        new_height = cast("Scalar", self.styles.height).value + delta

        if tail and new_height != cast("Scalar", self.styles.height).value:
            self.offset += Offset(0, -delta)

        self.styles.height = new_height
        return delta

    def set_dims(
        self,
        offset: int | None = None,
        size: int | None = None,
    ) -> None:
        if offset:
            self.offset = Offset(0, offset)
        if size:
            self.styles.height = size

    @property
    def start(self) -> int:
        return self.offset.y


class HorizontalEntry(AbstractEntry):
    """Horizontal entry for a horizontal timeline layout.

    Params:
        id: ID of the Widget.
        content: A Rich renderable, or string containing console markup.
        offset: Initial offset to use for the entry.
        size: Initial size to use for the entry.
        markup: True if markup should be parsed and rendered.
        name: Name of widget.
        classes: Space separated list of class names.
        disabled: Whether the static is disabled or not.
    """

    class Resized(AbstractEntry.Resized["HorizontalEntry"]):
        """Entry was resized by the user."""

    class Moved(AbstractEntry.Moved["HorizontalEntry"]):
        """Entry was moved by the user."""

    DIMENSION = "width"
    EDGE_MARGIN: ClassVar[int] = 2
    DEFAULT_CSS: ClassVar[str] = """\
    HorizontalEntry {
        min-width: 2;
        width: 4;
        height: 100%;
        border-left: double $panel-lighten-2;
        border-right: double $panel-lighten-2;
        &:hover.size_start {
            border: none;
            border-left: double $secondary;
            border-right: double $panel-lighten-2;
        }
        &:hover.size_end {
            border: none;
            border-right: double $secondary;
            border-left: double $panel-lighten-2;
        }
        &:focus.size_start {
            border: none;
            border-left: thick $primary;
            border-right: double $primary;
        }
        &:focus.size_end {
            border: none;
            border-left: double $primary;
            border-right: thick $primary;
        }
        &:focus {
            border: double $primary;
        }
        &.-mime {
            hatch: vertical white 5%;
        }
    }
    """
    """Default CSS for `HorizontalEntry` widget."""

    def sel_delta(self, event: MouseEvent) -> int:
        return event.delta_x

    def is_tail(self, offset: Offset) -> bool:
        return 0 <= offset.x <= self.EDGE_MARGIN

    def is_head(self, offset: Offset) -> bool:
        width = cast("Scalar", self.styles.width).value
        return width - self.EDGE_MARGIN <= offset.x <= width

    def is_start(self, offset: Offset) -> bool:
        return offset.x < cast("Scalar", self.styles.width).value / 2

    def move(self, delta: int) -> None:
        offset = Offset(delta, 0)
        self.offset = self.offset + offset
        self.post_message(self.Moved(self, delta))

    def merge(self, other: AbstractEntry) -> AwaitRemove:
        x1, x2 = self.offset.x, other.offset.x
        start = x1 if x1 <= x2 else x2
        self.offset = Offset(start, 0)
        e1, e2 = self.end, other.end
        self.styles.width = abs(start - (e1 if e1 >= e2 else e2))
        return other.remove()

    def _resize_helper(self, delta: int, *, tail: bool) -> int:
        if tail:
            delta *= -1

        new_width = cast("Scalar", self.styles.width).value + delta

        if tail and new_width != cast("Scalar", self.styles.width).value:
            self.offset += Offset(-delta, 0)

        self.styles.width = new_width
        return delta

    def set_dims(
        self, offset: int | None = None, size: int | None = None
    ) -> None:
        if offset:
            self.offset = Offset(offset, 0)
        if size:
            self.styles.width = size

    @property
    def start(self) -> int:
        return self.offset.x



"""Utilities for formatting time useful for when providing callables'."""

from __future__ import annotations

from ._utility import breakdown_seconds
from ._utility import format_seconds

__all__ = ("breakdown_seconds", "format_seconds")



from __future__ import annotations

import inspect
import sys
from functools import cached_property
from typing import TYPE_CHECKING
from typing import ClassVar
from typing import Generic
from typing import Iterator
from typing import TypeVar

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from rich.text import Text
from textual.dom import DOMNode
from textual.message import Message
from textual.reactive import reactive
from textual.reactive import var
from textual.widget import Widget
from textual.widgets import Button

from textual_timepiece.constants import LOCKED_ICON
from textual_timepiece.constants import TARGET_ICON
from textual_timepiece.constants import UNLOCKED_ICON

if TYPE_CHECKING:
    from rich.console import RenderableType
    from textual.app import RenderResult
    from textual.geometry import Offset


class BaseWidget(Widget):
    """Base class with a bunch of utility methods."""

    async def recompose(self) -> None:
        self.clear_cached_properties()
        await super().recompose()

    def clear_cached_properties(self) -> None:
        for prop in self.get_cached_properties():
            if hasattr(self, prop):
                delattr(self, prop)

    @classmethod
    def get_cached_properties(cls) -> Iterator[str]:
        for n, v in inspect.getmembers(cls):
            if isinstance(v, cached_property):
                yield n

    def get_line_offset(self, offset: Offset) -> str:
        x = offset.x - int(self.styles.border_left[0] != "")
        index = 0

        for seg in self.render_line(
            offset.y - self._top_border_offset()
        )._segments:
            # REFACTOR: Look for a public method for this.
            index += len(seg.text)
            if index > x:
                return str(seg.text.strip())

        return ""

    def _top_border_offset(self) -> int:
        return int(self.styles.border_top[0] != "")

    def disable(self, *, disable: bool = True) -> Self:
        self.disabled = disable
        return self

    def disable_messages(  # type: ignore[override]  # NOTE: Easier access to disabling in compose.
        self,
        *messages: type[Message],
    ) -> Self:
        super().disable_messages(*messages)
        return self

    def notify_style_update(self) -> None:
        # FIX: This is hacky way of dealing with rendering issues that
        # Textualize/textual@dca94393e77a9fa05df7b1f88865e1aec8d26054
        # introduced
        for child in self.query():
            child.notify_style_update()
        super().notify_style_update()


class LockButton(Button, BaseWidget):
    DEFAULT_CSS: ClassVar[str] = """\
    LockButton {
        # background: transparent;
        color: auto;
        min-width: 4;
        max-width: 4;
        height: 3;
        content-align-vertical: middle;
        text-style: bold;
        border: none;

        &:focus {
            background-tint: $primary 50%;
        }
    }
    """

    locked = var[bool](False, init=False)
    icon = reactive[Text](Text(LOCKED_ICON), init=False)

    def __init__(
        self,
        *,
        is_locked: bool = True,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        tooltip: RenderableType | None = None,
        action: str | None = None,
        use_variant: bool = False,
    ) -> None:
        self._use_variant = use_variant
        super().__init__(
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
            tooltip=tooltip,
            action=action,
        )
        self.locked = is_locked

    def compute_icon(self) -> Text:
        return Text(
            LOCKED_ICON if self.locked else UNLOCKED_ICON,
            self.rich_style,
        )

    def render(self) -> RenderResult:
        return self.icon

    def press(self) -> Self:
        self.locked = not self.locked
        return super().press()

    def watch_locked(self) -> None:
        if self._use_variant:
            self.variant = "warning" if self.locked else "success"


class ExpandButton(Button):
    """Button with a special icon."""

    DEFAULT_CSS: ClassVar[str] = """\
    ExpandButton {
        background: transparent;
        height: 3;
        min-width: 3;
        max-width: 3;
        border: none;
        padding: 0;
    }
    """

    expanded = var[bool](False, init=False)

    def __init__(
        self,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        *,
        disabled: bool = False,
        tooltip: RenderableType | None = None,
        action: str | None = None,
    ) -> None:
        super().__init__(
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
            tooltip=tooltip,
            action=action,
        )

    def render(self) -> RenderResult:
        return Text("▲" if self.expanded else "▼", self.rich_style)


class TargetButton(Button):
    def __init__(
        self,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        *,
        disabled: bool = False,
        tooltip: RenderableType | None = None,
        action: str | None = None,
    ) -> None:
        super().__init__(
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
            tooltip=tooltip,
            action=action,
        )

    def render(self) -> RenderResult:
        return Text(TARGET_ICON, self.rich_style)


WidgetType = TypeVar("WidgetType", bound=DOMNode)


class BaseMessage(Message, Generic[WidgetType]):
    """Generic message that overrides the control method."""

    def __init__(self, widget: WidgetType) -> None:
        super().__init__()
        self.widget = widget

    @property
    def control(self) -> WidgetType:
        return self.widget



# noqa: D100

from __future__ import annotations

__name__ = "textual-timepiece"
__description__ = "Various time related widgets & functionality for Textual."
__url__ = "https://github.com/ddkasa/textual-timepiece"
__author__ = "David Kasakaitis"
__author_email__ = "davidkasakaitis@proton.me"
__version__ = "0.6.0"
__license__ = "MIT"



"""Widgets for displaying & interacting with data on a timeline scale."""

from __future__ import annotations

from ._base_timeline import HorizontalRuler
from ._base_timeline import HorizontalTimeline
from ._base_timeline import HorizontalTimelineNavigation
from ._base_timeline import VerticalRuler
from ._base_timeline import VerticalTimeline
from ._base_timeline import VerticalTimelineNavigation
from ._timeline_entry import HorizontalEntry
from ._timeline_entry import VerticalEntry
from ._timeline_manager import RuledHorizontalTimeline
from ._timeline_manager import RuledVerticalTimeline

__all__ = (
    "HorizontalEntry",
    "HorizontalRuler",
    "HorizontalTimeline",
    "HorizontalTimelineNavigation",
    "RuledHorizontalTimeline",
    "RuledVerticalTimeline",
    "VerticalEntry",
    "VerticalRuler",
    "VerticalTimeline",
    "VerticalTimelineNavigation",
)



from __future__ import annotations

from textual.reactive import var

from textual import on, work
from textual.app import ComposeResult, App
from textual.containers import Center, Horizontal, HorizontalGroup, Middle
from textual.screen import ModalScreen
from textual.validation import Integer
from textual.widgets import Button, Input, Label, Static

from textual_timepiece.timeline import RuledHorizontalTimeline
from textual_timepiece.timeline import HorizontalTimeline


NUMBERS: tuple[str, ...] = [
    "Zero",
    "One",
    "Two",
    "Three",
    "Four",
    "Five",
    "Six",
    "Seven",
    "Eight",
    "Nine",
    "Ten",
]


def header_factory(index: int) -> Label:
    return Label(
        f"{NUMBERS[index]}",
        variant="primary" if index % 2 == 0 else "secondary",
        classes="header",
    )


class TimelineApp(App[None]):
    """Example of how horizontal timelines could be implemented."""

    DEFAULT_CSS = """\
    Screen {
        Input {
            width: 1fr;
        }
        HorizontalRuler {
            padding-left: 11;  # Compensating for header size.
        }
        Label.header {
            color: auto;
            border: hkey $secondary;
            min-width: 11;
            max-width: 11;
            height: 100%;
            text-align: center;
            content-align: center middle;
            text-style: bold;
            padding: 0 2 0 2;
        }
    }
    """

    layers = var[int](2, init=False)
    """Total amount of timelines to display."""

    def compose(self) -> ComposeResult:
        with HorizontalGroup(id="layer-controls"):
            yield Button.warning("-", id="subtract")
            yield Input(
                "2",
                "Total Timelines",
                type="integer",
                valid_empty=False,
                validate_on=["changed"],
                validators=Integer(1, 10),
                tooltip=(
                    "Total Timelines Present\n"
                    "[b]Maximum[/]: 10\n"
                    "[b]Minimum[/]: 1"
                ),
            )
            yield Button.success("+", id="add")
        yield (timeline := RuledHorizontalTimeline(
            3,
            header_factory=header_factory,
        ).data_bind(total=TimelineApp.layers))
        timeline.length = 392

    def _on_input_changed(self, message: Input.Changed) -> None:
        if message.input.is_valid:
            self.layers = int(message.value)

    def _on_button_pressed(self, message: Button.Pressed) -> None:
        self.layers -= 1 if message.button.id == "subtract" else -1

    def watch_layers(self, total: int) -> None:
        self.query_one("#subtract").disabled = total == 1
        self.query_one("#add").disabled = total == 10
        with (input_widget := self.query_one(Input)).prevent(Input.Changed):
            input_widget.value = str(total)


if __name__ == "__main__":
    TimelineApp().run()



import pytest
from textual.app import App

from textual_timepiece.timeline._timeline_entry import VerticalEntry


class TimelineApp(App[None]):
    def compose(self):
        yield VerticalEntry("first-entry", "Test Content", 5, 10)

    @property
    def entry(self) -> VerticalEntry:
        return self.query_one(VerticalEntry)


@pytest.fixture
def entry_app():
    return TimelineApp()


@pytest.mark.unit
async def test_entry_move(entry_app: TimelineApp):
    async with entry_app.run_test() as pilot:
        await pilot.pause()
        entry = entry_app.entry
        assert entry.offset.y == 5
        entry.move(5)
        assert entry.offset.y == 10


@pytest.mark.unit
async def test_entry_resize(entry_app: TimelineApp):
    async with entry_app.run_test() as pilot:
        await pilot.pause()
        entry = entry_app.entry
        assert entry.styles.height.value == 10
        entry.resize(2, tail=False)
        assert entry.styles.height.value == 12
        entry.resize(-2, tail=True)
        assert entry.styles.height.value == 14
        entry.resize(2, tail=True)
        assert entry.styles.height.value == 12


@pytest.mark.unit
async def test_entry_merge(entry_app: TimelineApp):
    async with entry_app.run_test() as pilot:
        entry = entry_app.entry

        new_entry = VerticalEntry("second-entry", "Test Content", 4, 16)
        await entry_app.mount(new_entry)
        await entry.merge(new_entry)
        await pilot.pause()

        assert entry.offset.y == 4
        assert entry.styles.height.value == 16



import pytest
from textual.app import App

from textual_timepiece.timeline._timeline_manager import (
    RuledHorizontalTimeline,
)
from textual_timepiece.timeline._timeline_manager import RuledVerticalTimeline


class TimelineManagerApp(App[None]):
    def __init__(self, t_type, **kwargs):
        super().__init__()
        self.t_type = t_type

    async def on_mount(self):
        entry = self.t_type.Timeline.Timeline.Entry
        total = 0
        for tl in self.manager.query("AbstractTimeline"):
            await tl.mount_all(
                [
                    entry(f"id-{i}", str(i), (i * 2) + total, 3)
                    for i in range(1, 4)
                ]
            )
            total += 5

    def compose(self):
        yield self.t_type()

    @property
    def manager(self):
        return self.query_one(self.t_type)


@pytest.fixture
def vert_tl_manager_app():
    return TimelineManagerApp(RuledVerticalTimeline)


@pytest.fixture
def horz_tl_manager_app():
    return TimelineManagerApp(RuledHorizontalTimeline)


@pytest.mark.snapshot
def test_vertical_manager_snapshot(vert_tl_manager_app, snap_compare):
    assert snap_compare(vert_tl_manager_app)


@pytest.mark.snapshot
def test_horizontal_manager_snapshot(horz_tl_manager_app, snap_compare):
    assert snap_compare(horz_tl_manager_app)



from __future__ import annotations

import sys
from abc import abstractmethod
from types import MappingProxyType
from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import ClassVar
from typing import Generic
from typing import TypeAlias
from typing import cast

if sys.version_info >= (3, 13):
    from typing import TypeVar
else:
    from typing_extensions import TypeVar

from rich.segment import Segment
from rich.style import Style as RichStyle
from textual.binding import Binding
from textual.binding import BindingType
from textual.css.query import NoMatches
from textual.css.query import QueryType
from textual.geometry import Offset
from textual.geometry import Region
from textual.geometry import Size
from textual.reactive import Reactive
from textual.reactive import reactive
from textual.reactive import var
from textual.strip import Strip
from textual.widget import AwaitMount
from textual.widget import Widget

from textual_timepiece._extra import BaseMessage
from textual_timepiece._utility import format_seconds

from ._timeline_entry import AbstractEntry
from ._timeline_entry import HorizontalEntry
from ._timeline_entry import VerticalEntry
from ._timeline_layouts import AbstractTimelineLayout
from ._timeline_layouts import HorizontalTimelineLayout
from ._timeline_layouts import VerticalTimelineLayout

if TYPE_CHECKING:
    from collections.abc import Iterable

    from textual.app import ComposeResult
    from textual.await_remove import AwaitRemove
    from textual.css.scalar import Scalar
    from textual.events import DescendantBlur
    from textual.events import DescendantFocus
    from textual.events import MouseDown
    from textual.events import MouseMove
    from textual.events import MouseUp

EntryType = TypeVar("EntryType", bound="AbstractEntry")
TimelineType = TypeVar("TimelineType", bound="AbstractTimeline[Any]")

EntryT = TypeVar("EntryT", bound="AbstractEntry")


class AbstractTimeline(Widget, Generic[EntryType], can_focus=True):
    """Abstract timeline implementation with various items.

    Describes a few abstract methods for creating entry with user input.
    The chain goes as follows:
        `_on_mouse_down` -> `_on_mouse_move` -> `_on_mouse_up`
        -> `calculate_entry_size` -> Post `EntryCreated` message.

    Params:
        *children: Entries to intially add to the widget.
        duration: Duration of size of the widget.
        name: The name of the widget.
        id: The ID of the widget in the DOM.
        classes: The CSS classes for the widget.
        disabled: Whether the widget is disabled or not.
        tile: Whether to tile the timeline or not.
    """

    class Updated(BaseMessage[TimelineType], Generic[EntryT, TimelineType]):
        """Base class for all timeline messages."""

        def __init__(self, widget: TimelineType, entry: EntryT) -> None:
            super().__init__(widget)

            self.entry = entry
            """Entry that was updated."""

        @property
        def timeline(self) -> TimelineType:
            """Alias for `widget` attribute."""
            return self.widget

    class Created(Updated[EntryT, TimelineType]):
        """Sent when a new entry is created."""

    class Deleted(Updated[EntryT, TimelineType]):
        """Sent when an entry is deleted."""

    class Selected(Updated[EntryT, TimelineType]):
        """Sent when a new entry selected."""

    Markers: TypeAlias = MappingProxyType[int, tuple[RichStyle, str]]
    Entry: type[EntryType]
    Layout: type[AbstractTimelineLayout[Any]]

    DURATION: ClassVar[str]
    BINDING_GROUP_TITLE: str = "Timeline"
    BINDINGS: ClassVar[list[BindingType]] = [
        Binding(
            "ctrl+down,ctrl+right",
            "adjust_tail",
            tooltip="Move entry to the backward.",
        ),
        Binding(
            "ctrl+up,ctrl+left",
            "adjust_head",
            tooltip="Move entry to the forward.",
        ),
        Binding(
            "alt+shift+down,alt+shift+left",
            "adjust_tail(True, True)",
            tooltip="Resize the tail end of the entry.",
        ),
        Binding(
            "alt+shift+up,alt+shift+right",
            "adjust_head(True, True)",
            tooltip="Resize the end of the entry forward.",
        ),
        Binding(
            "shift+up,shift+left",
            "adjust_head(False, True)",
            tooltip="Resize the start of the entry backward.",
        ),
        Binding(
            "shift+down,shift+right",
            "adjust_tail(False, True)",
            tooltip="Move the head of the entry forward.",
        ),
        Binding(
            "ctrl+d,delete,backspace",
            "delete_entry",
            "Delete Entry",
            tooltip="Delete the selected entry.",
        ),
        Binding(
            "escape",
            "clear_active",
            "Clear",
            priority=True,
            show=False,
            tooltip="Cancel creating an entry or deselect entries.",
        ),
    ]
    """All bindings for `AbstractTimeline` widget.

    | Key(s) | Description |
    | :- | :- |
    | ctrl+down,ctrl+right | Move entry to the backward. |
    | ctrl+up,ctrl+left | Move entry to the forward. |
    | alt+shift+down,alt+shift+left | Resize the tail end of the entry. |
    | alt+shift+up,alt+shift+right | Resize the end of the entry forward. |
    | shift+up,shift+left | Resize the start of the entry backward. |
    | shift+down,shift+right | Move the head of the entry forward. |
    | ctrl+d,delete,backspace | Delete the selected entry. |
    | escape | Cancel creating an entry or deselect entries. |
    """

    COMPONENT_CLASSES: ClassVar[set[str]] = {
        "timeline--normal",
    }
    """All component classes for the `AbstractTimeline` widget.

    | Class | Description |
    | :- | :- |
    | `timeline--normal` | Target all lines without set markers. |
    """
    DEFAULT_CSS: ClassVar[str] = """\
    AbstractTimeline {
        background: $panel-darken-1;
        .timeline--normal {
            color: white 5%;
        }
    }
    """
    """Default CSS for the `AbstractTimeline` widget."""

    children: list[EntryType]
    _start: Offset | None = None
    _mime: EntryType | None = None

    length = Reactive[int](96, init=False, layout=True)
    """Actual size of the widget with the direction size of the widget."""

    markers = Reactive[Markers](
        MappingProxyType({}),
        init=False,
        compute=False,
        repaint=False,
    )
    """Custom markers to place on the timeline."""

    def __init__(
        self,
        *children: EntryType,
        duration: int | None = None,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        tile: bool = True,
    ) -> None:
        super().__init__(
            *children,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
        )

        self._highlighted: EntryType | None = None
        # TODO: Convert to multiple selections.

        self._layout = self.Layout(tile=tile)
        if duration:
            self.length = duration

    async def _on_descendant_focus(self, event: DescendantFocus) -> None:
        self._highlighted = cast("EntryType", event.widget)
        self.post_message(self.Selected(self, self._highlighted))

    async def _on_descendant_blur(self, event: DescendantBlur) -> None:
        if event.widget is self._highlighted:
            self._highlighted = None

    @abstractmethod
    def _watch_markers(self, old: Markers, new: Markers) -> None: ...

    def mount(  # type: ignore[override] # NOTE: Making sure the user mounts the right widgets.
        self,
        *widgets: EntryType,
        before: str | int | EntryType | None = None,
        after: str | int | EntryType | None = None,
    ) -> AwaitMount:
        return super().mount(*widgets, before=before, after=after)

    def mount_all(  # type: ignore[override] # NOTE: Making sure the user mounts the right widgets.
        self,
        widgets: Iterable[EntryType],
        *,
        before: str | int | EntryType | None = None,
        after: str | int | EntryType | None = None,
    ) -> AwaitMount:
        return super().mount_all(widgets, before=before, after=after)

    async def _on_mouse_down(self, event: MouseDown) -> None:
        if self.app.mouse_captured is None:
            self.capture_mouse()
            self._start = event.offset
            if self._mime:
                self.post_message(self.Created(self, self._mime))
                self._mime = None

    async def _create_mime(self, offset: Offset) -> None:
        self._mime = self.Entry.mime(*self._calc_entry_size(offset))
        self.refresh_bindings()

        await self.mount(self._mime)

    async def _on_mouse_move(self, event: MouseMove) -> None:
        if self._start is None or self.app.mouse_captured != self:
            return

        if self._mime is None:
            await self._create_mime(event.offset)
        else:
            self._mime.set_dims(*self._calc_entry_size(event.offset))

    async def _on_mouse_up(self, event: MouseUp) -> None:
        if self.app.mouse_captured == self:
            self.capture_mouse(False)
        if self._mime:
            self.post_message(self.Created(self, self._mime))
            self._start = None
            self._mime = None

    def remove_children(
        self,
        selector: str | type[QueryType] | Iterable[EntryType] = "*",  # type: ignore[override] # NOTE: Type should always be an AbstractEntry
    ) -> AwaitRemove:
        return super().remove_children(selector)

    def check_action(
        self,
        action: str,
        parameters: tuple[object, ...],
    ) -> bool | None:
        if action == "clear_active":
            return self._mime is not None or self.selected is not None
        if action in {"adjust_tail", "adjust_head"}:
            return self.selected is not None

        return True

    def action_delete_entry(self, id: str | None = None) -> None:
        """Remove the selected or provided entry from the timeline.

        Args:
            id: If removing an un-highlighted widget.
        """
        if id:
            try:
                entry = self.query_one(f"#{id}", self.Entry)
            except NoMatches:
                return
        elif self.selected:
            entry = self.selected
        else:
            return

        entry.remove()
        self.post_message(self.Deleted(self, entry))

    def _action_clear_active(self) -> None:
        if self._mime:
            self._mime.remove()
            self._start = None
            self._mime = None
        else:
            cast("EntryType", self.selected).blur()

    def action_adjust_tail(
        self,
        tail: bool = False,
        resize: bool = False,
    ) -> None:
        """Adjust the tail of the selected timeline entry.

        Args:
            tail: Increase the size if resizing.
            resize: Resize the entry instead of moving.
        """
        if resize:
            cast("EntryType", self.selected).resize(1, tail=tail)
        else:
            cast("EntryType", self.selected).move(1)

    def action_adjust_head(
        self,
        tail: bool = False,
        resize: bool = False,
    ) -> None:
        """Adjust the head of the selected timeline entry.

        Args:
            tail: Increase the size if resizing.
            resize: Resize the entry instead of moving.
        """
        if resize:
            cast("EntryType", self.selected).resize(-1, tail=not tail)
        else:
            cast("EntryType", self.selected).move(-1)

    @abstractmethod
    def _calc_entry_size(self, end: Offset) -> tuple[int, int]:
        """Calculate the size of an entry based off the offsets.

        Assumes that self._start is not None.

        Args:
            end: Offset were this method was called.

        Returns:
            A start point and size of the new widget.
        """

    def _watch_duration(self, value: int) -> None:
        setattr(self.styles, self.DURATION, value)

    @property
    def layout(self) -> AbstractTimelineLayout[EntryType]:
        return self._layout

    @property
    def tile(self) -> bool:
        """Is calendar tiling enabled?"""
        return self._layout.tile

    @tile.setter
    def tile(self, value: bool) -> None:
        self._layout.tile = value
        self.refresh(layout=True)

    @property
    def selected(self) -> EntryType | None:
        """Currently highlighted entry. None if there is nothing selected."""
        return self._highlighted


VerticalEntryType = TypeVar(
    "VerticalEntryType",
    bound=VerticalEntry,
    default=VerticalEntry,
)

VerticalEntryT = TypeVar(
    "VerticalEntryT",
    bound=VerticalEntry,
    default=VerticalEntry,
)

VerticalTimelineType = TypeVar(
    "VerticalTimelineType",
    bound="VerticalTimeline",
    default="VerticalTimeline",
)


class VerticalTimeline(AbstractTimeline[VerticalEntryType]):
    """Timeline that displays entries in a vertical view.

    Params:
        *children: Entries to intially add to the widget.
        duration: Size of the widget.
        name: The name of the widget.
        id: The ID of the widget in the DOM.
        classes: The CSS classes for the widget.
        disabled: Whether the widget is disabled or not.
        tile: Whether to tile the timeline or not. Set to `False` if a single
            column of entries is preferred.
    """

    Entry = VerticalEntry  # type: ignore[assignment] # FIX: Need to research to how to correctly accomplish this.
    Layout = VerticalTimelineLayout
    DURATION = "height"
    DEFAULT_CSS: ClassVar[str] = """\
    VerticalTimeline {
        height: auto !important;
        border-left: wide $secondary;
        border-right: wide $secondary;
        &:hover {
            border-left: thick $secondary;
            border-right: thick $secondary;
        }
        &:focus {
            border-left: outer $primary;
            border-right: outer $primary;
        }
    }
    """
    """Default CSS for `VerticalTimeline` widget."""

    class Created(
        AbstractTimeline.Created[VerticalEntryT, VerticalTimelineType]
    ):
        """Sent when a new entry is created."""

    class Deleted(
        AbstractTimeline.Deleted[VerticalEntryT, VerticalTimelineType]
    ):
        """Sent when an entry is deleted."""

    class Selected(
        AbstractTimeline.Selected[VerticalEntryT, VerticalTimelineType]
    ):
        """Sent when a new entry selected."""

    def _watch_markers(
        self,
        old: AbstractTimeline.Markers,
        new: AbstractTimeline.Markers,
    ) -> None:
        for line in old.keys() ^ new.keys():
            self.refresh_line(line)

    def refresh_line(self, y: int) -> None:
        """Refresh a single line.

        Args:
            y: Coordinate of line.
        """
        self.refresh(
            Region(
                0,
                y - self.scroll_offset.y,
                max(self.virtual_size.width, self.size.width),
                1,
            )
        )

    def render_lines(self, crop: Region) -> list[Strip]:
        self._basic_strip = Strip(
            [
                Segment(
                    "─" * self.size.width,
                    style=self.get_component_rich_style("timeline--normal"),
                )
            ]
        )
        return super().render_lines(crop)

    def render_line(self, y: int) -> Strip:
        if marker := self.markers.get(y):
            style, label = marker

            return Strip(
                [Segment(label.center(self.size.width, "─"), style=style)]
            )

        return self._basic_strip

    def _calc_entry_size(self, end: Offset) -> tuple[int, int]:
        start = cast("Offset", self._start)
        return start.y if start.y < end.y else end.y, abs(end.y - start.y)

    def pre_layout(self, layout: VerticalTimelineLayout) -> None:  # type: ignore[override]
        # Sorting with the public sort method will cause infinite layout calls.
        self._nodes._sort(  # TODO: This needs a public solution.
            key=lambda w: (w.offset.y, cast("Scalar", w.styles.height).value),
        )

    def get_content_height(
        self,
        container: Size,
        viewport: Size,
        width: int,
    ) -> int:
        return self.length


HorizontalEntryType = TypeVar(
    "HorizontalEntryType",
    bound=HorizontalEntry,
    default=HorizontalEntry,
)

HorizontalEntryT = TypeVar(
    "HorizontalEntryT",
    bound=HorizontalEntry,
    default=HorizontalEntry,
)
HorizontalTimelineType = TypeVar(
    "HorizontalTimelineType",
    bound="HorizontalTimeline",
    default="HorizontalTimeline",
)


class HorizontalTimeline(AbstractTimeline[HorizontalEntryType]):
    """Timeline that displays entries in a horizontal view.

    Params:
        *children: Entries to intially add to the widget.
        duration: Size of the widget.
        name: The name of the widget.
        id: The ID of the widget in the DOM.
        classes: The CSS classes for the widget.
        disabled: Whether the widget is disabled or not.
        tile: Whether to tile the timeline or not. Set to `False` if a single
            row of entries is preferred.
    """

    Entry = HorizontalEntry  # type: ignore[assignment] # FIX: Need to research to how to correctly accomplish this.
    Layout = HorizontalTimelineLayout
    DURATION: ClassVar[str] = "width"
    DEFAULT_CSS: ClassVar[str] = """\
    HorizontalTimeline {
        width: auto !important;
        height: 28;
        border-top: tall $secondary;
        border-bottom: tall $secondary;
        &:hover {
            border-top: thick $secondary;
            border-bottom: thick $secondary;
        }
        &:focus {
            border-top: outer $primary;
            border-bottom: outer $primary;
        }
    }
    """
    """Default CSS for `HorizontalTimeline` widget."""

    class Created(
        AbstractTimeline.Created[HorizontalEntryT, HorizontalTimelineType]
    ):
        """Sent when a new entry is created."""

    class Deleted(
        AbstractTimeline.Deleted[HorizontalEntryT, HorizontalTimelineType]
    ):
        """Sent when an entry is deleted."""

    class Selected(
        AbstractTimeline.Selected[HorizontalEntryT, HorizontalTimelineType]
    ):
        """Sent when a new entry selected."""

    def _create_strip(self) -> Strip:
        """Prerenders the strip for reuse on each line."""
        defaults = (self.get_component_rich_style("timeline--normal"), "")

        segs = list[Segment]()
        add_seg = segs.append
        prev_style = None
        current_strip = ""
        for x in range(self.size.width):
            style, _ = self.markers.get(x, defaults)
            if prev_style and style != prev_style:
                add_seg(Segment(current_strip, prev_style))
                current_strip = ""

            prev_style = style
            current_strip += "│"

        add_seg(Segment(current_strip, prev_style))

        return Strip(segs, self.size.width).simplify()

    def render_lines(self, crop: Region) -> list[Strip]:
        self._cached_strip = self._create_strip()
        return super().render_lines(crop)

    def render_line(self, y: int) -> Strip:
        return self._cached_strip

    def _watch_markers(
        self,
        old: AbstractTimeline.Markers,
        new: AbstractTimeline.Markers,
    ) -> None:
        if old.keys() ^ new.keys():
            self.refresh()

    def _calc_entry_size(self, end: Offset) -> tuple[int, int]:
        start = cast("Offset", self._start)
        return start.x if start.x < end.x else end.x, abs(end.x - start.x)

    def pre_layout(self, layout: HorizontalTimelineLayout) -> None:  # type: ignore[override]
        # Sorting with the public sort method will cause infinite layout calls.
        self._nodes._sort(  # TODO: This needs a public solution .
            key=lambda w: (w.offset.x, cast("Scalar", w.styles.width).value),
        )

    def get_content_width(self, container: Size, viewport: Size) -> int:
        return self.length


class AbstractRuler(Widget):
    """Abstract ruler class for marking timelines with custom markers.

    Params:
        duration: Total length of the ruler.
        marker_factory: Callable for creating the markers.
        name: The name of the widget.
        id: The ID of the widget in the DOM.
        classes: The CSS classes for the widget.
        disabled: Whether the widget is disabled or not.
    """

    MarkerFactory: TypeAlias = Callable[[int], str]

    COMPONENT_CLASSES: ClassVar[set[str]] = {
        "abstractruler--label",
        "abstractruler--empty",
    }
    DEFAULT_CSS: ClassVar[str] = """\
    AbstractRuler {
        background: $panel-darken-3;

        .abstractruler--label {
            color: $secondary;
            text-style: italic;
        }
        .abstractruler--empty {
            color: $panel 50%;
        }
    }
    """
    """Default CSS for the `AbstractRuler` widget."""

    duration = var[int](86400, init=False)
    """Total time actual time the ruler spans in seconds."""

    length = reactive[int](96, layout=True, init=False)
    """Actual length of the widget."""

    subdivisions = reactive[int](24, init=False)
    """Amount of subdivisions to use when calculating markers.

    Generator gets called this amount of times.
    """

    time_chunk = Reactive[int](3600, init=False, compute=False)
    """Time chunk for each subdivision that the ruler creates.

    Computed automatically when other reactives change.
    """

    marker_len = Reactive[int](4, init=False, compute=False)
    """The marker length of the each time_chunk."""

    def __init__(
        self,
        duration: int | None = None,
        subdivisions: int | None = None,
        marker_factory: MarkerFactory | None = None,
        *,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ) -> None:
        super().__init__(
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
            markup=False,
        )
        if duration:
            self.length = duration
        if subdivisions:
            self.subdivisions = subdivisions

        self._factory = marker_factory or format_seconds

    def _compute_time_chunk(self) -> int:
        return self.duration // self.subdivisions

    def _compute_marker_len(self) -> int:
        return self.length // self.subdivisions


class VerticalRuler(AbstractRuler):
    """Vertical ruler for marking vertical timelines.

    Params:
        duration: Total length of the ruler.
        marker_factory: Callable for creating the markers.
        name: The name of the widget.
        id: The ID of the widget in the DOM.
        classes: The CSS classes for the widget.
        disabled: Whether the widget is disabled or not.
    """

    DEFAULT_CSS: ClassVar[str] = """\
    VerticalRuler {
        border-left: wide $secondary;
        border-right: wide $secondary;
        height: auto !important;
        width: 8;
    }
    """
    """Default CSS for the `VerticalRuler` widget."""

    def render_line(self, y: int) -> Strip:
        marker_pos, rem = divmod(y, self.marker_len)
        if y and not rem:
            return Strip(
                [
                    Segment(
                        self._factory(marker_pos * self.time_chunk).center(
                            self.size.width
                        ),
                        style=self.get_component_rich_style(
                            "abstractruler--label"
                        ),
                    )
                ]
            )

        return Strip(
            [
                Segment(
                    f" {'─' * (self.size.width - 2)} ",
                    style=self.get_component_rich_style(
                        "abstractruler--empty"
                    ),
                )
            ]
        )

    def get_content_height(
        self,
        container: Size,
        viewport: Size,
        width: int,
    ) -> int:
        return self.length


class HorizontalRuler(AbstractRuler):
    """Horizontal ruler for marking horizontal timelines.

    Params:
        duration: Total length of the ruler.
        marker_factory: Callable for creating the markers.
        name: The name of the widget.
        id: The ID of the widget in the DOM.
        classes: The CSS classes for the widget.
        disabled: Whether the widget is disabled or not.
    """

    DEFAULT_CSS: ClassVar[str] = """\
    HorizontalRuler {
        border-top: tall $secondary;
        border-bottom: tall $secondary;
        width: auto !important;
        height: 3;
        hatch: vertical white 5%;
    }
    """
    """Default CSS for `HorizontalRuler` widget."""

    def render_line(self, y: int) -> Strip:
        if y != (self.size.height // 2):
            return Strip.blank(self.size.width)

        style = self.get_component_rich_style("abstractruler--label")
        return Strip(
            [
                Segment(self._factory(t).rjust(self.marker_len), style)
                for t in range(self.time_chunk, self.duration, self.time_chunk)
            ]
        ).simplify()

    def get_content_width(self, container: Size, viewport: Size) -> int:
        return self.length


ChildTimeline = TypeVar("ChildTimeline", bound="AbstractTimeline[Any]")


class TimelineNavigation(Widget, Generic[ChildTimeline]):
    """Container Widget for a single timline and its header.

    Params:
        header: Header to use at the start of the timeline.
        name: The name of the widget.
        id: The ID of the widget in the DOM.
        classes: The CSS classes for the widget.
        disabled: Whether the widget is disabled or not.
    """

    Timeline: type[ChildTimeline]

    length = var[int](96, init=False)
    """Actual length of the actual timeline navigation."""

    def __init__(
        self,
        header: Widget | None = None,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        *,
        disabled: bool = False,
    ) -> None:
        super().__init__(
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
            markup=False,
        )

        self._header = header
        if header:
            header.add_class("-timeline-header")

    def compose(self) -> ComposeResult:
        if self._header:
            yield self._header
        yield self.Timeline().data_bind(TimelineNavigation.length)

    @property
    def timeline(self) -> ChildTimeline:
        return self.query_exactly_one(self.Timeline)


class VerticalTimelineNavigation(
    TimelineNavigation[VerticalTimeline[VerticalEntryType]]
):
    """Vertical widget containing a vertical timeline and header.

    Params:
        header: Header to use at the start of the timeline.
        name: The name of the widget.
        id: The ID of the widget in the DOM.
        classes: The CSS classes for the widget.
        disabled: Whether the widget is disabled or not.
    """

    Timeline = VerticalTimeline[VerticalEntryT]

    DEFAULT_CSS: ClassVar[str] = """\
    VerticalTimelineNavigation {
        layout: vertical !important;
        height: auto !important;
    }
    """
    """Default CSS for `VerticalTimelineNavigation` widget."""


class HorizontalTimelineNavigation(
    TimelineNavigation[HorizontalTimeline[HorizontalEntryType]]
):
    """Horizontal widget containing a horizontal timeline and header.

    Params:
        header: Header to use at the start of the timeline.
        name: The name of the widget.
        id: The ID of the widget in the DOM.
        classes: The CSS classes for the widget.
        disabled: Whether the widget is disabled or not.
    """

    Timeline = HorizontalTimeline[HorizontalEntryT]

    DEFAULT_CSS: ClassVar[str] = """\
    HorizontalTimelineNavigation {
        layout: horizontal !important;
        width: auto !important;
    }
    """
    """Default CSS for `HorizontalTimelineNavigation` widget."""



from __future__ import annotations

import inspect
import random
from collections import defaultdict
from dataclasses import dataclass
from functools import cached_property
from typing import TYPE_CHECKING
from typing import ClassVar
from typing import Literal

from rich.pretty import Pretty
from rich.syntax import Syntax
from textual import on
from textual.app import App
from textual.app import ComposeResult
from textual.containers import Container
from textual.containers import Horizontal
from textual.containers import ScrollableContainer
from textual.events import Mount
from textual.message import Message
from textual.screen import ModalScreen
from textual.widget import Widget
from textual.widgets import Button
from textual.widgets import Footer
from textual.widgets import Header
from textual.widgets import Label
from textual.widgets import Static
from textual.widgets import TabbedContent
from textual.widgets import TabPane

from textual_timepiece.__about__ import __version__
from textual_timepiece._activity_heatmap import ActivityHeatmap
from textual_timepiece._activity_heatmap import HeatmapManager
from textual_timepiece.pickers import DatePicker
from textual_timepiece.pickers import DateRangePicker
from textual_timepiece.pickers import DateTimeDurationPicker
from textual_timepiece.pickers import DateTimePicker
from textual_timepiece.pickers import DateTimeRangePicker
from textual_timepiece.pickers import DurationPicker
from textual_timepiece.pickers import TimePicker
from textual_timepiece.pickers._date_picker import DateSelect
from textual_timepiece.pickers._time_picker import DurationSelect
from textual_timepiece.pickers._time_picker import TimeSelect
from textual_timepiece.timeline._timeline_manager import (
    RuledHorizontalTimeline,
)
from textual_timepiece.timeline._timeline_manager import RuledVerticalTimeline

if TYPE_CHECKING:
    from rich.console import RenderableType
    from textual.events import Mount


class DemoWidget(Widget):
    """Displays each widget with additional information."""

    @dataclass
    class Toggle(Message):
        """Sent when the user presses one of the buttons."""

        widget: type[Widget]
        preview: Literal[
            "docstring", "tcss", "code", "docs", "source", "bindings"
        ]

    def __init__(
        self,
        widget_call: type[Widget],
        *,
        notes: str | None = None,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ) -> None:
        super().__init__(name=name, id=id, classes=classes, disabled=disabled)

        self._notes = notes
        self._widget_type = widget_call

    def _compose_navigation_bar(self) -> ComposeResult:
        with Horizontal(id="navigation"):
            yield Button("Docstring", id="docstring", classes="nav")
            yield Label(self._widget_type.__name__, classes="title")
            if (
                hasattr(self._widget_type, "DEFAULT_CSS")
                and self._widget_type.DEFAULT_CSS
            ):
                yield Button("Default CSS", id="default-css", classes="nav")
            if (
                hasattr(self._widget_type, "BINDINGS")
                and self._widget_type.BINDINGS
            ):
                yield Button("Bindings", id="bindings", classes="nav")

            yield Button("Code Preview", id="code", classes="nav")

    def compose(self) -> ComposeResult:
        """Compose the layout for the widget."""
        yield from self._compose_navigation_bar()
        yield self._widget_type()

    @on(Button.Pressed, "#docstring")
    def _open_docstring(self, message: Button.Pressed) -> None:
        self.post_message(self.Toggle(self._widget_type, "docstring"))

    @on(Button.Pressed, "#default-css")
    def _open_default_css(self, message: Button.Pressed) -> None:
        self.post_message(self.Toggle(self._widget_type, "tcss"))

    @on(Button.Pressed, "#code")
    def _open_source(self, message: Button.Pressed) -> None:
        self.post_message(self.Toggle(self._widget_type, "code"))

    @on(Button.Pressed, "#bindings")
    def _open_bindings(self, message: Button.Pressed) -> None:
        self.post_message(self.Toggle(self._widget_type, "bindings"))


class PreviewScreen(ModalScreen[None]):
    """Preview screen that pops up to display associated widget info."""

    BINDINGS: ClassVar = [
        ("escape", "hide_preview", "Close Preview"),
    ]

    def __init__(
        self,
        renderable: RenderableType,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
    ) -> None:
        super().__init__(name, id, classes)
        self._renderable = renderable

    def _on_mount(self, event: Mount) -> None:
        self.refresh_bindings()

    def compose(self) -> ComposeResult:
        """Generate the layout for the screen."""
        with Container():
            with ScrollableContainer():
                yield Static(self._renderable, id="preview")

            with Horizontal():
                yield Button(
                    r"Quit\[esc]",
                    "warning",
                    action="screen.hide_preview",
                    classes="nav",
                )

    def _action_hide_preview(self) -> None:
        self.dismiss()


class TimepieceDemo(App[None]):
    """Main demo app that displays the widgets."""

    CSS: ClassVar[str] = """
    Screen {
        layout: horizontal;
        align: center middle;

        TabbedContent {
            padding: 1 0;
            min-width: 135;
            width: 65%;

            TabPane {
                height: auto;
            }

            Container.previews {
                layout: vertical;
                overflow-y: auto;
                content-align-horizontal: center;
                align-horizontal: center;
            }
        }
    }

    .title {
        background: $panel;
        content-align-horizontal: center;
        text-align: center;
        width: 1fr;
        text-style: bold;
    }

    DemoWidget {
        height: auto;
        margin: 2 0;
        padding: 0 2;

        Horizontal#navigation {
            height: auto;
            margin-bottom: 1;
        }
    }

    Button.nav {
        border: none;
        height: 1;
        width: 10%;
        min-width: 13;
    }

    .widget_container {
        height: auto;
        width: 100%;
        align: center middle;
    }

    PreviewScreen {
        align: center middle;

        Container {
            border: round $primary;
            width: 50%;
            min-width: 125;
            height: 90%;

            ScrollableContainer {
                Static#preview {
                    margin: 1 2;
                }
            }

            Horizontal {
                height: 1;
                margin-top: 1;
                align-horizontal: center;
            }

        }
    }
    """

    TITLE = "Textual Timepiece"
    SUB_TITLE = __version__

    def compose(self) -> ComposeResult:
        """Generate the main layout for the demo app."""
        yield Header(show_clock=True)

        with TabbedContent(initial="pickers"):
            with (
                TabPane("Pickers", id="pickers"),
                Container(id="Pickers", classes="previews"),
            ):
                for item in (
                    DatePicker,
                    DurationPicker,
                    TimePicker,
                    DateTimePicker,
                    DateRangePicker,
                    DateTimeRangePicker,
                    DateTimeDurationPicker,
                ):
                    yield DemoWidget(item)

            with (
                TabPane("Select"),
                Container(id="Pickers", classes="previews"),
            ):
                for select in (
                    DateSelect,
                    TimeSelect,
                    DurationSelect,
                ):
                    yield DemoWidget(select)

            with (
                TabPane("Heatmap"),
                Container(id="heatmap", classes="previews"),
            ):
                for i in (ActivityHeatmap, HeatmapManager):
                    yield DemoWidget(i)

            with (
                TabPane("Timeline"),
                Container(id="timeline", classes="previews"),
            ):
                for timeline in (
                    RuledVerticalTimeline,
                    RuledHorizontalTimeline,
                ):
                    yield DemoWidget(timeline)

        yield Footer()

    def _on_mount(self, event: Mount) -> None:
        for widget in self.query(ActivityHeatmap):
            self._set_data(widget)

    @on(DemoWidget.Toggle)
    def _open_tab(self, message: DemoWidget.Toggle) -> None:
        data: RenderableType
        if message.preview == "tcss":
            data = message.widget.DEFAULT_CSS

        elif message.preview == "bindings":
            data = Pretty(message.widget.BINDINGS)

        elif message.preview == "docstring":
            data = str(message.widget.__doc__)

        elif message.preview == "code":
            data = Syntax(
                inspect.getsource(message.widget),
                "python",
                line_numbers=True,
                padding=1,
            )

        self.app.push_screen(PreviewScreen(data))

    def _set_data(self, widget: ActivityHeatmap) -> None:
        random.seed(widget.year)
        template = ActivityHeatmap.generate_empty_activity(widget.year)
        widget.values = defaultdict(
            lambda: 0,
            {
                day: random.randint(6000, 20000)  # noqa: S311
                for week in template
                for day in week
                if day
            },
        )

    @on(HeatmapManager.YearChanged)
    def _change_heat_year(self, message: HeatmapManager.YearChanged) -> None:
        message.stop()
        self._set_data(message.widget.heatmap)

    @cached_property
    def preview_panel(self) -> Static:
        """Panel for displaying information about a widget."""
        return self.query_one("#preview", Static)


def main() -> None:
    """Main entry point for the demo application."""
    TimepieceDemo().run()


if __name__ == "__main__":
    main()



from datetime import datetime

import pytest

from textual_timepiece._utility import time_to_seconds
from textual_timepiece.utility import breakdown_seconds
from textual_timepiece.utility import format_seconds


@pytest.mark.unit
def test_time_to_seconds():
    dt = time_to_seconds(datetime(2025, 4, 2, 12, 30, 30))
    h, m, s = breakdown_seconds(dt)
    assert h == 12
    assert m == 30
    assert s == 30


@pytest.mark.unit
def test_format_seconds():
    dt = time_to_seconds(datetime(2025, 4, 2, 12, 30, 30))
    assert format_seconds(dt) == "12:30:30"
    assert format_seconds(dt, include_seconds=False) == "12:30"



import pytest
from textual.app import App
from textual.css.query import NoMatches

from textual_timepiece.timeline._base_timeline import HorizontalTimeline
from textual_timepiece.timeline._base_timeline import VerticalTimeline


class TimelineApp(App[None]):
    def __init__(self, t_type, **kwargs):
        super().__init__()
        self.t_type = t_type

    async def on_mount(self):
        entry = self.t_type.Entry
        await self.timeline.mount_all(
            [entry(f"id-{i}", str(i), (i * 2), 3) for i in range(1, 5)]
        )

    def compose(self):
        yield self.t_type()

    @property
    def timeline(self):
        return self.query_one(self.t_type)


@pytest.fixture
def vert_tl_app():
    return TimelineApp(VerticalTimeline)


@pytest.fixture
def horz_tl_app():
    return TimelineApp(HorizontalTimeline)


@pytest.mark.snapshot
def test_vert_timeline_snapshot(vert_tl_app, snap_compare):
    assert snap_compare(vert_tl_app)


@pytest.mark.snapshot
def test_vert_timeline_no_tile_snapshot(vert_tl_app, snap_compare):
    async def run_before(pilot):
        pilot.app.timeline.tile = False
        await pilot.pause()

    assert snap_compare(vert_tl_app, run_before=run_before)


@pytest.mark.snapshot
def test_horz_timeline_snapshot(horz_tl_app, snap_compare):
    assert snap_compare(horz_tl_app)


@pytest.mark.snapshot
def test_horz_timeline_no_tile_snapshot(horz_tl_app, snap_compare):
    async def run_before(pilot):
        pilot.app.timeline.tile = False
        await pilot.pause()

    assert snap_compare(horz_tl_app, run_before=run_before)


@pytest.mark.unit
async def test_delete_entry(horz_tl_app):
    async with horz_tl_app.run_test() as pilot:
        horz_tl_app.timeline.action_delete_entry("id-3")
        await pilot.pause()

        with pytest.raises(NoMatches):
            horz_tl_app.timeline.query_one("#id-3")

        horz_tl_app.timeline.query_one("#id-2").focus()
        await pilot.pause()

        horz_tl_app.timeline.action_delete_entry()
        await pilot.pause()

        with pytest.raises(NoMatches):
            horz_tl_app.timeline.query_one("#id-2")

        horz_tl_app.timeline.query_one("#id-1").focus()
        await pilot.press("delete")

        with pytest.raises(NoMatches):
            horz_tl_app.timeline.query_one("#id-1")


@pytest.mark.unit
async def test_timeline_move_entry(horz_tl_app):
    async with horz_tl_app.run_test() as pilot:
        entry = horz_tl_app.timeline.query_one("#id-2").focus()
        current_end = entry.end
        await pilot.pause()

        await pilot.press("ctrl+right")
        assert entry.end == current_end + 1

        await pilot.press("ctrl+left")
        assert entry.end == current_end


@pytest.mark.unit
async def test_timeline_resize_entry(horz_tl_app):
    async with horz_tl_app.run_test() as pilot:
        entry = horz_tl_app.timeline.query_one("#id-2").focus()
        current_start = entry.start
        current_end = entry.end

        await pilot.press("shift+up")
        assert entry.start == current_start - 1

        await pilot.press("alt+shift+down")
        assert entry.start == current_start

        await pilot.press("shift+down")
        assert entry.end == current_end + 1

        await pilot.press("alt+shift+up")
        assert entry.end == current_end



from __future__ import annotations

import nox

nox.options.default_venv_backend = "uv|virtualenv"


@nox.session(python=["3.10"])
def lint(session: nox.Session) -> None:
    session.run("uv", "sync", "--active", "--no-dev", "--group", "lint")
    session.run("uv", "run", "--active", "ruff", "check", "src")


@nox.session(python=["3.10"])
def type_check(session: nox.Session) -> None:
    session.run("uv", "sync", "--active", "--no-dev", "--group", "type")
    session.run("uv", "run", "--active", "mypy", "src")


@nox.session(python=["3.10", "3.11", "3.12", "3.13", "3.14"])
def test(session: nox.Session) -> None:
    python_version = f"--python={session.python}"
    session.run(
        "uv",
        "sync",
        "--active",
        python_version,
        "--no-dev",
        "--group",
        "test",
    )
    session.run(
        "uv",
        "run",
        "--active",
        python_version,
        "pytest",
        "--cov-branch",
        "--cov-report=xml",
        "-n",
        "auto",
    )



from __future__ import annotations

from textual import on, work
from textual.app import ComposeResult, App
from textual.containers import Center, Middle
from textual.screen import ModalScreen
from textual.widgets import Input

from textual_timepiece.timeline import RuledVerticalTimeline
from textual_timepiece.timeline import VerticalTimeline


class NamingModal(ModalScreen[str]):
    """Modal screen for naming entries."""

    DEFAULT_CSS = """\
    NamingModal {
        align: center middle;
        Center{
            border: tall $primary;
            border-top: panel $primary;
            border-title-align: center;
            border-title-style: bold;
            min-width: 34;
            width: 50%;
            min-height: 5;
            height: 20%;
        }
        Input {
            min-width: 30;
        }
    }
    """
    BINDINGS = [("escape", "dismiss")]

    def compose(self) -> ComposeResult:
        with Center() as center, Middle():
            center.border_title = "What would you like to name the entry?"
            yield Input(
                placeholder="Name",
                valid_empty=False,
                validate_on=["submitted"],
            )

    def on_input_submitted(self, message: Input.Submitted) -> None:
        message.stop()
        if message.input.is_valid:
            self.dismiss(message.value)


class TimelineApp(App[None]):
    """Example of how timelines could be implemented."""

    def compose(self) -> ComposeResult:
        yield RuledVerticalTimeline(3)

    @work(name="naming-worker")
    async def on_vertical_timeline_created(
        self,
        message: VerticalTimeline.Created,
    ) -> None:
        result = await self.push_screen_wait(NamingModal())
        message.entry.remove_class("-mime").update(result or "To Be Named")

    def on_vertical_timeline_deleted(
        self,
        message: VerticalTimeline.Deleted,
    ) -> None:
        self.notify(f"Successfully deleted {message.entry.visual!s}!")


if __name__ == "__main__":
    TimelineApp().run()



import random
from collections import defaultdict

from textual.app import App, ComposeResult
from textual_timepiece.activity_heatmap import ActivityHeatmap, HeatmapManager


class ActivityApp(App[None]):
    def _on_heatmap_manager_year_changed(
        self,
        message: HeatmapManager.YearChanged,
    ) -> None:
        message.stop()
        self.set_heatmap_data(message.year)

    def retrieve_data(self, year: int) -> ActivityHeatmap.ActivityData:
        """Placeholder example on how the data could be generated."""
        random.seed(year)
        template = ActivityHeatmap.generate_empty_activity(year)
        return defaultdict(
            lambda: 0,
            {
                day: random.randint(6000, 20000)
                for week in template
                for day in week
                if day
            },
        )

    def set_heatmap_data(self, year: int) -> None:
        """Sets the data based on the current data."""
        self.query_one(ActivityHeatmap).values = self.retrieve_data(year)

    def _on_mount(self) -> None:
        self.set_heatmap_data(2025)

    def compose(self) -> ComposeResult:
        yield HeatmapManager(2025)


if __name__ == "__main__":
    ActivityApp().run()
